#!/usr/bin/env python
import sys, os, random
from cards import *


#This software is coypright 2008 by Thomas Dickerson
#and StickFigure Graphic Productions
#under the terms of the GNU General Public License v3
#a copy of which may be obtained here: http://www.gnu.org/licenses/gpl-3.0.txt

STAND = 0
DRAW = 1
TRADE = 2

BOMBOUT = -24
PURESABACC = 24
IDIOTSARRAY = 25

class SabaccCard(Card):
	def __init__(self, suit, value, deck):
		Card.__init__(self, suit, value, deck)
		self.active = True
		
	def __init__(self, suit, value, suitdict = None, valset = None):
		Card.__init__(self, suit, value, suitdict = suitdict, valset = valset)
		self.active = True
	
	def activate(self):
		self.active = True
		
	def neutralize(self):
		self.active = False
		
	def isactive(self):
		return self.active
	
	def __str__(self):
		if self.suit.lower() == "face":
			return self.value
		else:
			return "%s of %s" % (self.value, self.suit)

class SabaccPlayer(Player):
	def __init__(self, game=None, hand=None, name="", funds=0, motivator=None, observer=False):
		Player.__init__(self, game=game, hand=hand, name=name, observer=observer)
		self.funds = funds
		self.motivator = None
	
	def acceptfunds(self, amount):
		self.funds += amount
		
	def neutralizehand(self):
		for card in self.hand:
			card.neutralize()
		
	def getactivatedcards(self):
		return [card for card in self.hand if card.isactive()]
		
	def getneutralizedcards(self):
		return [card for card in self.hand if not card.isactive()]
		
	def paytopool(self, amount, sabaccpool=False):
		self.funds -= amount
		self.game.acceptintopool(amount, sabaccpool=sabaccpool)
		
	def pollcall(self):
		return (random.randint(0, 5) <= 2)
		
	def pollbetorfold(self, minbet):
		if minbet >= (self.funds * 0.75) or self.funds < 20:
			return -1
		elif minbet >= (self.funds * 0.5):
			return minbet
		else:
			return minbet + random.randint(0, 5)
		
	def polldrawtradestand(self):
		hand = self.gethand()
		return (random.randint(STAND, TRADE), hand[random.randint(0, len(hand) - 1)])
		
	def tradecard(self, card, deck):
		self.removecard(card)
		deck.addcard(card)
		deck.shuffle()
		self.acceptcard(deck.drawcard())
		
	def drawnewcard(self, deck):
		self.acceptcard(deck.drawcard())
		
	def pollneutralfieldexchange(self):
		incard = None
		activecards = self.getactivatedcards()
		outcard = None
		inactivecards = self.getneutralizedcards()
		for card in activecards:
			if random.randint(0, 5) <= 2:
				incard = card
				break
		for card in inactivecards:
			if random.randint(0, 5) <= 2:
				outcard = card
				break
		return (incard, outcard)
		
	def uptonshifts(self, n):
		activecards = self.getactivatedcards()
		deck = self.game.getdeck()
		for i in range(n):
			r = random.randint(0, len(activecards) - 1)
			card = activecards[i]
			activecards[i] = deck.drawrandomcard()
			deck.addcard(card)
			deck.shuffle()
			
	def handtostr(self):
		return " ".join([("---, " if card.isactive() else ("%s, " % str(card))) for card in self.gethand()])
		
	def __str__(self):
		return "Name: %s\tCredits: %d" % (self.name, self.funds)
		
			
class SabaccGame(Game):
	def __init__(self, mode=0, initialmpool=0, initialspool=0):
		modes = ['Standard Sabacc', 'Bespin Standard', 'Empress Teta Preferred', 'Cloud City Casino', 'Corellian Gambit', 'Wild Fives', 'Random Sabacc', 'Force Sabacc']
		Game.__init__(self, gametype = modes[mode])
		
		self.mainpool = initialmpool
		self.sabaccpool = initialspool
		
		self.normalsuits = {'Coins' : 'gold', 'Flasks' : 'gold', 'Sabers' : 'black', 'Staves' : 'black'}
		self.normalvalues = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', 'Commander', 'Mistress', 'Master', 'Ace']
		self.normallookup = [ 1,   2,   3,   4,   5,   6,   7,   8,   9,   10,   11,   12,          13,         14,       15  ]

		self.othersuits = {'Face' : 'black'}
		self.othervalues = ['Queen of Air and Darkness', 'Endurance', 'Balance', 'Demise', 'Moderation', 'The Evil One', 'The Star', 'The Idiot']
		self.otherlookup = [ -2,                          -8,          -11,       -13,      -14,          -15,            -17,        0         ]


		self.fullsuits = self.normalsuits.copy()
		for key, value in self.othersuits.iteritems():
			self.fullsuits[key] = value
		self.fullvalues = self.normalvalues + self.othervalues

		self.scoringtable = {}
		self.fulllookup = self.normallookup + self.otherlookup
		for i in range(0, len(self.fullvalues)):
			self.scoringtable[self.fullvalues[i]] = self.fulllookup[i]

		newdeck = []
		for suit in self.normalsuits.keys():
			for value in self.normalvalues:
				newdeck.append(SabaccCard(suit, value, suitdict = self.fullsuits, valset = self.fullvalues))
		for suit in self.othersuits.keys():
			for value in self.othervalues:
				newdeck.append(SabaccCard(suit, value, suitdict = self.fullsuits, valset = self.fullvalues))
				newdeck.append(SabaccCard(suit, value, suitdict = self.fullsuits, valset = self.fullvalues))
	
		self.setdeck(Deck(suitdict = self.fullsuits, valset = self.fullvalues, mergedecks = [newdeck]))
		self.getdeck().shuffle()
	
	def acceptintopool(self, amount, sabaccpool=False):
		if sabaccpool:
			self.sabaccpool += amount
		else:
			self.mainpool += amount
	
	def payoutfrompool(self, amount, sabaccpool=False):
		if sabaccpool:
			self.sabaccpool -= amount
		else:
			self.mainpool -= amount
		return amount
	
	def cancallgame(self, turncount, playercount):
		return turncount >= playercount
	
	def pollcallgame(self, players, whoseturn):
		for i in range(len(players)):
			if i != whoseturn and players[i].pollcall():
				return True
		return False
			
	def gamestatus(self):
		gametype, deck, players = self.observe()
		print "Gametype: '%s'" % gametype
		print "%d cards left in deck" % deck.countdeck()
		print ""
		print "Players: "
		for player in players:
			print "\t%s" % str(player)
			print "\tCards: %s" % player.handtostr()
			print ""
			print ""
			
	def doshift(self, players):
		for player in players:
			d1 = random.randint(1, 6)
			d2 = random.randint(1, 6)
			shifts = 0
			if d1 == d2: #(1 / 6)
				shifts += 1
				if not d1 % 2: #(1 / 12)
					shifts += 1
					if not d1 % 6: #(1 / 36)
						shifts += 1
			player.uptonshifts(shifts)
					
	
	def dobet(self, whoseturn):
		minbet = 0
		skipturn = False
		toremove = []
		for player in self.getplayers():
			betorfold = player.pollbetorfold(minbet)
			if betorfold < minbet:
				player.paytopool(1, sabaccpool=True)
				toremove.append(player)
				if player == self.getplayers()[whoseturn]:
					skipturn = True	
			else:
				minbet = betorfold
				player.paytopool(betorfold)
		for player in toremove:
			self.removeplayer(player)
		return skipturn
				
	def cycleturn(self, whoseturn):
		whoseturn += 1
		if whoseturn >= self.countplayers():
			whoseturn = 0
		return whoseturn
	
	def revealcards(self, players):
		for player in players:
			player.neutralizehand()
		self.gamestatus()
		
	def scorehand(self, hand):
		global BOMBOUT, PURESABACC, IDIOTSARRAY
		values = [self.scoringtable[card.getvalue()] for card in hand]
		values.sort()
		score = sum(values)
		specialhand = None
		relativevalue = score
		if score < -23 or score > 23 or score == 0:
			specialhand = "Bombed Out"
			relativevalue = BOMBOUT
		if score == 23:
			specialhand =  "Pure Sabacc"
			relativevalue = PURESABACC
		if len(values) >= 3 and (0 in values and 2 in values and 3 in values):
			specialhand = "Idiot's Array"
			relativevalue = IDIOTSARRAY
		
		return (relativevalue, score, specialhand)
		
	def paybombfee(self, player, total):
		print "%s bombed out with a total of %d and must contribute %d Credits to the Sabacc Pool." % (player.getname(), total, self.mainpool)
		player.paytopool(self.mainpool, sabaccpool=True)
		
	def handlenwaytie(self, n, players, rankings, totals, special):
		global BOMBOUT, PURESABACC, IDIOTSARRAY
		deck = self.getdeck()
		nscores = []
		# deal everyone a new card
		for i in range(-n, 0):
			players[i].drawnewcard(deck)
			rankings[i], totals[i], special[i] = self.scorehand(players[i].gethand())
			nscores.append((rankings[i], totals[i], special[i], player[i]))
		nrankings, ntotals, nspecial, nplayers = [list(tuple) for tuple in zip(*sorted(nscores))]
		#if we have a winner, then win
		highscore = max(nrankings)
		#if all winners bombed out take next best
		if highscore == BOMBOUT:
			scores = zip(rankings, totals, special, players)
			rankings, totals, special, players = [list(tuple) for tuple in zip(*sorted(scores))]
			nbhighscore = max(rankings)
			#if everyone bombed out return all bomb out
			if nbhighscore == BOMBOUT:
				return "All Players Bombed Out."
			else:
				pass
		#if we have a tie for first, repeat tie handling
		
		
	def handlewinner(self, player, rank, total, special):
		global PURESABACC, IDIOTSARRAY
		name = player.getname()
		mainpool = self.mainpool
		player.acceptfunds(self.payoutfrompool(mainpool))
		winstr = "%s wins the Main Pool (%d Credits) with a total of %d." % (name, mainpool, total)
		if rank in (PURESABACC, IDIOTSARRAY):
			sabaccpool = self.sabaccpool
			player.acceptfunds(self.payoutfrompool(sabaccpool, sabaccpool=True))
			winstr += "  %s also won the Sabacc Pool (%d Credits) with %s %s" % (name, sabaccpool, ("a" if rank == PURESABACC else "an"), special)
		return winstr
		
	def doscore(self, players):
		global BOMBOUT, PURESABACC, IDIOTSARRAY
		scores = []
		for player in players:
			hand = player.gethand()
			name = player.getname()
			relative, score, special = self.scorehand(hand)
			scores.append((relative, score, special, player))
			print str(player),
			print "\tScore: %s" % (special if special else str(score))
		rankings, actual, special, players = [list(tuple) for tuple in zip(*sorted(scores))]
		last = min(rankings)
		if last == BOMBOUT:
			numbombed = rankings.count(last)
			for player, total in zip(players[:numbombed], actual[:numbombed]):
				self.paybombfee(player, total)
			if numbombed == len(players):
				return "All Players Bombed Out."
		first = max(rankings)
		numfirst = rankings.count(first)
		if numfirst == 1:
			return self.handlewinner(players[-1], rankings[-1], actual[-1], special[-1])
		else:
			return self.handlenwaytie(numfirst, players, rankings, actual, special)

	def play(self):
		global TRADE, DRAW, STAND
		self.deal(toeach = 2)
		
		called = False
		whoseturn = 0
		turncount = 0
		self.deal(toeach = 2)
		while not called:
			gametype, deck, players = self.observe()
			self.gamestatus()
		
			turncount += 1
			skipturn = self.dobet(whoseturn)
			players = self.getplayers()
			playercount = self.countplayers()
			self.doshift(players)
			if skipturn:
				if playercount > 1:
					whoseturn = self.cycleturn(whoseturn)
					continue
				else:
					self.doscore(players)
					break
			if self.cancallgame(turncount, playercount) and self.pollcallgame(players, whoseturn):
				called = True
				self.revealcards(players)
				print self.doscore(players)
			else:
				action, card = players[whoseturn].polldrawtradestand()
				if action == TRADE:
					#trade card? discard or shuffle in?
					players[whoseturn].tradecard(card, deck)
				elif action == DRAW:
					players[whoseturn].drawnewcard(deck)
				else:
					pass
			
				incard, outcard = players[whoseturn].pollneutralfieldexchange()
				if incard:
					incard.neutralize()
				if outcard:
					outcard.activate()
				
				#cycle turn
				whoseturn = self.cycleturn(whoseturn)
	
mygame = SabaccGame()
#make 4 new players
players = [ SabaccPlayer(game = mygame, name = "Bob", funds=200), SabaccPlayer(game = mygame, name = "Fred", funds=300), SabaccPlayer(game = mygame, name = "Sally", funds=250), SabaccPlayer(game = mygame, name = "Harold", funds=400)]
mygame.setplayers(playergroup=players)
mygame.play()
